///////////////////Software Control Header///////////////////////////////////
// TTree.h - Library to support binary tree operation
//
// $LogThisFile$ \Projects\UtilCode\Include\TTree.h
// 
// V01.02 29/04/2001 1409
//  Accept pointer insertion.
// 
// V01.01 15/10/2000 2302
//  CTree - option to allow multiple insertion.
//  CTreeLooper - option to sort Left->Right & Right<-Left.
// 
// V01.00 10/10/2000 2159
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////

#ifndef __TS_TREE__
#define __TS_TREE__

#include <iostream.h>
#include "TSingLst.h"

/////////////////////////////////////////////////////////////////////////////
// CTreeNode - A single node of tree.
//
// Desctiption: A left node within tree.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> class CTreeNode
{
	T *m_data;
	CTreeNode<T> *m_left, *m_right;

public:
   CTreeNode();
	CTreeNode(T &, CTreeNode<T> *left = NULL, CTreeNode<T> *right = NULL);
	CTreeNode(T *, CTreeNode<T> *left = NULL, CTreeNode<T> *right = NULL);
   CTreeNode(CTreeNode<T> &sourceNode);   // Copy constructor
	~CTreeNode();

   ////////////////
   // Modifiers
   ////////////////

   T *Insert(T &, bool MultipleIns);
   T *Insert(T *, bool MultipleIns);
   CTreeNode<T> *Insert(CTreeNode<T> *thisNode, bool MultipleIns);
   // Set left node.
   void SetLeft(CTreeNode<T> *left) {m_left = left;}
   // Set right node.
   void SetRight(CTreeNode<T> *right) {m_right = right;}
   CTreeNode &operator=(CTreeNode<T> &sourceNode);

   ////////////////
   // Accessors
   ////////////////

   // Return data pointer.
   T *Data() {return m_data;}
   // Return left pointer.
   CTreeNode<T> *Left() {return m_left;}
   // Return right pointer.
   CTreeNode<T> *Right() {return m_right;}
};

/////////////////////////////////////////////////////////////////////////////
// CTreeNode()
//
// Desctiption: Constructor to create a default node.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeNode<T>::CTreeNode()
{
   m_data = new T;
   m_left = m_right = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// CTreeNode(CTreeNode<T> &sourceNode)
//
// Desctiption: Constructor to create an image of sourceNode.
//
// sourceNode - original node.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeNode<T>::CTreeNode(CTreeNode<T> &sourceNode)
{
   m_data = new T(*(sourceNode.m_data));
   m_left = m_right = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// operator=(CTreeNode<T> &sourceNode)
//
// Desctiption: Constructor to replace this node's data with sourceNode.
//
// sourceNode - original node.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeNode<T> &CTreeNode<T>
   ::operator=(CTreeNode<T> &sourceNode)
{
   if(m_data)
      delete m_data;
   m_data = new T(*(sourceNode.m_data));

   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// CTreeNode(T &data, CTreeNode<T> *left, CTreeNode<T> *right)
//
// Desctiption: Constructor to create a node.
//
// data - original copy.
// left - left pointer.
// right - right pointer.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeNode<T>::CTreeNode(T &data, CTreeNode<T> *left,
	CTreeNode<T> *right)
{
	m_data = new T(data);
	m_left = left;
	m_right = right;
}

/////////////////////////////////////////////////////////////////////////////
// CTreeNode(T *data, CTreeNode<T> *left, CTreeNode<T> *right)
//
// Desctiption: Constructor to create a node.
//
// data - Pointer to the inserted data.
// left - Left pointer.
// right - Right pointer.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeNode<T>::CTreeNode(T *data, CTreeNode<T> *left,
	CTreeNode<T> *right)
{
	m_data = data;
	m_left = left;
	m_right = right;
}

/////////////////////////////////////////////////////////////////////////////
// ~CTreeNode()
//
// Desctiption: Cleanup.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeNode<T>::~CTreeNode()
{
	if(m_data)
		delete m_data;
}

/////////////////////////////////////////////////////////////////////////////
// Insert(T &element, bool MultipleIns)
//
// Desctiption: Insert element into approriate left location.
//
// element - Original object.
// MultipleIns - True to allow multiple insertion.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CTreeNode<T>::Insert(T &element, bool MultipleIns)
{
   if(element < *m_data)
   {
      if(m_left)
         return m_left->Insert(element, MultipleIns);
      else
      {
         m_left = new CTreeNode<T>(element);
         return m_left->Data();
      }
   }
   // Insert at right for larger element
   // Allow same data if m_MultipleIns is true
   else if(element > *m_data || MultipleIns)
   {
      if(m_right)
         return m_right->Insert(element, MultipleIns);
      else
      {
         m_right = new CTreeNode<T>(element);
         return m_right->Data();
      }
   }

   return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Insert(T *element, bool MultipleIns)
//
// Desctiption: Insert element into approriate left location.
//
// element - Pointer to the inserted element.
// MultipleIns - True to allow multiple insertion.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CTreeNode<T>::Insert(T *element, bool MultipleIns)
{
   if(*element < *m_data)
   {
      if(m_left)
         return m_left->Insert(element, MultipleIns);
      else
      {
         m_left = new CTreeNode<T>(element);
         return m_left->Data();
      }
   }
   // Insert at right for larger element
   // Allow same data if m_MultipleIns is true
   else if(*element > *m_data || MultipleIns)
   {
      if(m_right)
         return m_right->Insert(element, MultipleIns);
      else
      {
         m_right = new CTreeNode<T>(element);
         return m_right->Data();
      }
   }

   return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Insert(CTreeNode<T> *thisNode)
//
// Desctiption: Insert a node.
//
// thisNode - Source node.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeNode<T> *CTreeNode<T>::Insert(CTreeNode<T> *thisNode,
                                            bool MultipleIns)
{
   if(thisNode->Data() < *m_data)
   {
      if(m_left)
         m_left->Insert(thisNode, MultipleIns);
      else
         m_left = thisNode;

      return thisNode;
   }
   else if(thisNode->Data() > *m_data || MultipleIns)
   {
      if(m_right)
         m_right->Insert(thisNode, MultipleIns);
      else
         m_right = thisNode;

      return thisNode;
   }

   return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// CTree - Implementation of binary tree.
//
// TSTan   19 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> class CTree
{
	CTreeNode<T> *m_root;
   bool m_MultipleIns, m_InvalidSize;
	int m_Size;

   void RemoveSubTree(CTreeNode<T> *);
   CTreeNode<T> *Remove(CTreeNode<T> *, T &, bool &);
   T *FindNode(T &, CTreeNode<T> *);
	int SizeLooping(CTreeNode<T> *Node, int Count);

public:
	CTree(bool multiple = true);
   CTree(CTree<T> &sourceTree);
	~CTree();

   ////////////////
   // Modifiers
   ////////////////

	int CalcSize();
   void SetInsertion(bool Multiple) {m_MultipleIns = Multiple;}
   T *Insert(T &);
   T *Insert(T *);
   void InsertTree(CTreeNode<T> *startNode);
	void InsertTree(CTreeNode<T> &startNode);
   bool Remove(T &);
	CTreeNode<T> *Remove(T &, bool CleanMemory);
   void RemoveAll();
   CTree<T> &operator=(CTree<T> &sourceTree);
   CTree<T> &operator=(CTreeNode<T> *rootNode);

   ////////////////
   // Accessors
   ////////////////

   // Return pointer to node element.
   T *GetNode(T &element) {return FindNode(element, m_root);}
   // Return root.
   CTreeNode<T> *Root() {return m_root;}
	int Size();
};

/////////////////////////////////////////////////////////////////////////////
// CalcSize()
//
// Desctiption: Return total elements in the tree with recalculation.
//
// TSTan   29 Apr 2001
/////////////////////////////////////////////////////////////////////////////
template<class T> int CTree<T>::CalcSize()
{
	m_InvalidSize = false;
	m_Size = 0;

	m_Size = SizeLooping(m_root, m_Size);

	return m_Size;
}

/////////////////////////////////////////////////////////////////////////////
// RemoveAll()
//
// Desctiption: Remove all nodes.
//
// TSTan   19 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> void CTree<T>::RemoveAll()
{
   RemoveSubTree(m_root);
   m_root = NULL;
	m_Size = 0;
}

/////////////////////////////////////////////////////////////////////////////
// CTree(CTree<T> &sourceTree)
//
// Desctiption: Construct an image from sourceTree.
//
// TSTan   19 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTree<T>::CTree(CTree<T> &sourceTree)
{
   m_root = NULL;
	m_Size = 0;
   m_MultipleIns = sourceTree.m_MultipleIns;
   InsertTree(sourceTree.m_root);
}

/////////////////////////////////////////////////////////////////////////////
// InsertTree(CTreeNode<T> *startNode)
//
// Desctiption: Insert subtree from startNode into tree.
//
// TSTan   19 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> void CTree<T>::InsertTree(CTreeNode<T> *startNode)
{
   if(startNode)
   {
      Insert(startNode->Data());
      InsertTree(startNode->Left());
      InsertTree(startNode->Right());
   }
}

/////////////////////////////////////////////////////////////////////////////
// InsertTree(CTreeNode<T> &startNode)
//
// Desctiption: Insert subtree from startNode into tree.
//
// TSTan   21 Mac 2001
/////////////////////////////////////////////////////////////////////////////
template<class T> void CTree<T>::InsertTree(CTreeNode<T> &startNode)
{
   if(startNode)
   {
      Insert(*(startNode.Data()));
      InsertTree(startNode.Left());
      InsertTree(startNode.Right());
   }
}

/////////////////////////////////////////////////////////////////////////////
// operator=(CTree<T> &sourceTree)
//
// Desctiption: Replace with the image of sourceTree.
//
// TSTan   19 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTree<T> &CTree<T>::operator=(CTree<T> &sourceTree)
{
   RemoveAll();
   m_root = NULL;
	m_Size = 0;
   m_MultipleIns = sourceTree.m_MultipleIns;
   InsertTree(sourceTree.m_root);

   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// operator=(CTreeNode<T> *rootNode)
//
// Desctiption: Replace with the image of sourceTree.
//
// TSTan   15 Oct 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTree<T> &CTree<T>::operator=(CTreeNode<T> *rootNode)
{
   RemoveAll();
   m_root = rootNode;

   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// CTree()
//
// Desctiption: Creates an empty tree.
//
// TSTan   19 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTree<T>::CTree(bool multiple)
{
   m_MultipleIns = multiple;
	m_root = NULL;
	m_Size = 0;
	m_InvalidSize = false;
}

/////////////////////////////////////////////////////////////////////////////
// ~CTree()
//
// Desctiption: Cleanup.
//
// TSTan   19 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTree<T>::~CTree()
{
   RemoveSubTree(m_root);
}

/////////////////////////////////////////////////////////////////////////////
// RemoveSubTree(CTreeNode<T> *nodePtr)
//
// Desctiption: Remove tree starting from nodePtr.
//
// nodePtr - starting node.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> void CTree<T>::RemoveSubTree(CTreeNode<T> *nodePtr)
{
   if(!m_root)
      return;

   if(nodePtr->Left())
      RemoveSubTree(nodePtr->Left());
   if(nodePtr->Right())
      RemoveSubTree(nodePtr->Right());
   delete nodePtr;
	m_Size--;
}

/////////////////////////////////////////////////////////////////////////////
// Remove(T &element, bool CleanMemory)
//
// Desctiption: Remove a node from tree.
//
// element - element to be removed from tree.
// CleanMemory - true to clean memory.
//
// TSTan   20 Mac 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeNode<T> *CTree<T>::Remove(T &element, bool CleanMemory)
{
	CTreeNode<T> *pTreeNode = m_root;
	CTreeNode<T> *ParentNode = NULL;

	while(1)
	{
		if(!pTreeNode)	// Node does not exist
			return NULL;

		if(*pTreeNode->Data() > element)
		{
			ParentNode = pTreeNode;
			pTreeNode = pTreeNode->Left();
		}
		else if(*pTreeNode->Data() < element)
		{
			ParentNode = pTreeNode;
			pTreeNode = pTreeNode->Right();
		}
		else		// found node
			break;
	}

	if(ParentNode)	// Not root
	{
		if(ParentNode->Left() == pTreeNode)
			ParentNode->SetLeft(NULL);
		else
			ParentNode->SetRight(NULL);

		InsertTree(pTreeNode->Left());
		InsertTree(pTreeNode->Right());
	}
	else	// Root node
	{
		m_root = pTreeNode->Left();
		InsertTree(pTreeNode->Right());
	}

	if(CleanMemory)	// clean the node
	{
		delete pTreeNode;
		return NULL;
	}
	m_InvalidSize = true;
	
	return pTreeNode;
}

/////////////////////////////////////////////////////////////////////////////
// Remove(T &element)
//
// Desctiption: Remove a node from tree.
//
// element - element to be removed from tree.
//
// TSTan   19 Aug 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> bool CTree<T>::Remove(T &element)
{
   bool Found = false;
   m_root = Remove(m_root, element, Found);
   return Found;
}

/////////////////////////////////////////////////////////////////////////////
// Remove(CTreeNode<T> *thisNode, T &target, bool &IsDeleted)
//
// Desctiption: Recursive function to quickly remove a node from tree.
//
// thisNode - current node.
// target - element to be removed.
// IsDeleted - true if element found and deleted.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeNode<T> *CTree<T>::Remove(CTreeNode<T> *thisNode,
   T &target, bool &IsDeleted)
{
   if(thisNode)
   {
      if(target > thisNode->Data())
         thisNode->SetRight(Remove(thisNode->Right(), target, IsDeleted));
      else if(target < thisNode->Data())
         thisNode->SetLeft(Remove(thisNode->Left(), target, IsDeleted));
      else
      {
         CTreeNode<T> *newSubTree;

         if(thisNode->Left())
         {
            thisNode->Left()->Insert(thisNode->Right());
            newSubTree = thisNode->Left();
         }
         else
            newSubTree = thisNode->Right();

         delete thisNode;
         IsDeleted = true;
			m_Size--;
         return newSubTree;
      }

      return thisNode;
   }

   return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Size()
//
// Desctiption: Return total elements int the tree.
//
// TSTan   29 Apr 2001
/////////////////////////////////////////////////////////////////////////////
template<class T> int CTree<T>::Size()
{
	if(m_InvalidSize)
		return CalcSize();

	return m_Size;
}

/////////////////////////////////////////////////////////////////////////////
// SizeLooping(CTreeNode<T> *Node, int Count)
//
// Desctiption: Use recursively to return total elements.
//
// Node - start node.
// Count - current node count;
//
// TSTan   29 Apr 2001
/////////////////////////////////////////////////////////////////////////////
template<class T> int CTree<T>::SizeLooping(CTreeNode<T> *Node, int Count)
{
	if(!Node)
		return Count;

	if(Node->Left())
		Count = SizeLooping(Node->Left(), Count);

	if(Node->Right())
		Count = SizeLooping(Node->Right(), Count);

	return Count + 1;
}

/////////////////////////////////////////////////////////////////////////////
// FindNode(T &element, CTreeNode<T> *thisNode)
//
// Desctiption: Check the existence of element within tree.
//
// element - element to search for.
// thisNode - starting node.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CTree<T>::FindNode(T &element,
   CTreeNode<T> *thisNode)
{
   if(thisNode)
   {
      if(element > *thisNode->Data())
         return FindNode(element, thisNode->Right());
      else if(element < *thisNode->Data())
         return FindNode(element, thisNode->Left());
      else
         return thisNode->Data();
   }

   return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Insert(T &element)
//
// Desctiption: Insert element to the tree.
//
// element - original element.
//
// TSTan   19 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CTree<T>::Insert(T &element)
{
	if(m_root)
	{
		T *pElement = m_root->Insert(element, m_MultipleIns);
		
		if(pElement)
			m_Size++;

		return pElement;
	}
	else
   {
		m_root = new CTreeNode<T>(element);
		m_Size++;
      return m_root->Data();
   }
}

/////////////////////////////////////////////////////////////////////////////
// Insert(T *element)
//
// Desctiption: Insert element to the tree.
//
// element - Pointer to inserted element.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CTree<T>::Insert(T *element)
{
	if(m_root)
	{
		T *pElement = m_root->Insert(element, m_MultipleIns);

		if(pElement)
			m_Size++;

		return pElement;
	}
	else
   {
		m_root = new CTreeNode<T>(element);
      return m_root->Data();
   }
}

/////////////////////////////////////////////////////////////////////////////
// CTreeLooper - Retrieve all elements within a given tree.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> class CTreeLooper
{
   CSinglyList<T*> m_allElements;
   bool m_SortOrder; // true - ascending, false - descending

   void GetElements(CTreeNode<T> *theTreeNode);

public:
   CTreeLooper(CTree<T> &theTree, bool SortOrder = true);
   CTreeLooper(CTreeLooper<T> &theLoop);
   ~CTreeLooper();

   ////////////////
   // Modifiers
   ////////////////

   void Reset();
   // Sort right->left(true) or left->right(false)
   void SetSortOrder(bool SortOrder) {m_SortOrder = SortOrder;}
   void NewTree(CTree<T> &theTree);
   CTreeLooper &operator=(CTreeLooper<T> &source);
   void operator++(int);   //postfix operator
   
   ////////////////
   // Accessors
   ////////////////

   T *Current();
   bool IsValid();
};

/////////////////////////////////////////////////////////////////////////////
// CTreeLooper(CTreeLooper<T> &source)
//
// Desctiption: Copy constructor.
//
// source - the other looper.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeLooper<T>::CTreeLooper(CTreeLooper<T> &source)
{
   m_allElements = source.m_allElements;
   m_SortOrder = source.m_SortOrder;
}

/////////////////////////////////////////////////////////////////////////////
// CTreeLooper(CTree<T> &theTree)
//
// Desctiption: Constructor with given tree.
//
// theTree - source tree to retrieve elements.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeLooper<T>::CTreeLooper(CTree<T> &theTree,
															 bool SortOrder)
{
   m_SortOrder = SortOrder;
   GetElements(theTree.Root());
   Reset();
}

/////////////////////////////////////////////////////////////////////////////
// ~CTreeLooper()
//
// Desctiption: Cleanup.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeLooper<T>::~CTreeLooper()
{}

/////////////////////////////////////////////////////////////////////////////
// Reset()
//
// Desctiption: Reset pointer to the first element.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> void CTreeLooper<T>::Reset()
{
   m_allElements.ResetPtr();
}

/////////////////////////////////////////////////////////////////////////////
// NewTree(CTree<T> &theTree)
//
// Desctiption: Assign new tree to the looper.
//
// theTree - the new tree.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> void CTreeLooper<T>::NewTree(CTree<T> &theTree)
{
   m_allElements.RemoveAll();
   GetElements(theTree.Root());
   Reset();
}

/////////////////////////////////////////////////////////////////////////////
// operator=(CTreeLooper<T> &source)
//
// Desctiption: Assignment operation.
//
// source = the other looper.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> CTreeLooper<T> &CTreeLooper<T>::
   operator=(CTreeLooper<T> &source)
{
   m_SortOrder = source.m_SortOrder;
   m_allElements = source.m_allElements;
   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// Current()
//
// Desctiption: Return pointer to current element of tree.
//              Modify original copy if content is changed.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> T *CTreeLooper<T>::Current()
{
   return *m_allElements.Current();
}

/////////////////////////////////////////////////////////////////////////////
// operator++(int)
//
// Desctiption: Posfix incremenet operator to point to next element.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> void CTreeLooper<T>::operator++(int)
{
   m_allElements.Next();
}

/////////////////////////////////////////////////////////////////////////////
// IsValid()
//
// Desctiption: True if current element is valid.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> bool CTreeLooper<T>::IsValid()
{
   return m_allElements.IsValid();
}

/////////////////////////////////////////////////////////////////////////////
// GetElements(CTreeNode<T> *theTreeNode)
//
// Desctiption: Recursive function to collect all elements into singly list.
//
// theTreeNode - the starting node.
//
// TSTan   24 Sep 2000
/////////////////////////////////////////////////////////////////////////////
template<class T> void CTreeLooper<T>::GetElements(CTreeNode<T> *theTreeNode)
{
   if(theTreeNode)
   {
      GetElements(theTreeNode->Left());
      T *FInfoPtr = theTreeNode->Data();
      m_allElements.Insert(FInfoPtr,
         m_SortOrder? CSinglyList<T*>::HEAD: CSinglyList<T*>::TAIL);
      GetElements(theTreeNode->Right());
   }
}

#endif




